package sysmodels.coordinategeometry;

import java.util.*;
import javax.media.j3d.*;
import javax.vecmath.*;
import com.sun.j3d.utils.geometry.*;

import core.*;
import util.*;

class CoordinateGeometry extends SystemModel
{
	CellProgram cellProgram;
	Vector cellDetails;
	int currCellID;
	Hashtable statistics;

	public CoordinateGeometry(Simulation p_sim)
	{
		super(p_sim);
		currCellID = 0;
	}

	public void init(Vector args)
	{
		String modelOptions = (String) args.elementAt(0);
		cellProgram = getSimulation().getCellProgram();
		cellDetails = new Vector();
		statistics = new Hashtable();
	}

	public void runCellProgramForOneStep()
	{
		Enumeration enum = cellDetails.elements();
		while(enum.hasMoreElements())
		{
			System.out.println("Reached Here");
			CellDetails cd = (CellDetails)enum.nextElement();
			Cell cell = cd.getCell();
			Vector argVector = new Vector();
			argVector.add(new String("Emit Chemicals"));
			cell.provideService(argVector);
		}

		enum = cellDetails.elements();
		while(enum.hasMoreElements())
		{
			CellDetails cd = (CellDetails)enum.nextElement();
			if(cd.isNew())
			{
				cd.setAsOld();
			}
			else
			{
				Cell cell = cd.getCell();
				cell.runCellProgramForOneStep();
			}
		}

		/* Now distribute the chemicals diffused during the run to individual cells */
		enum = cellDetails.elements();
		while(enum.hasMoreElements())
		{
			Vector argVector = new Vector();
			CellDetails cd = (CellDetails)enum.nextElement();
			Cell cell = cd.getCell();
			argVector.add(new String("Sense Chemicals"));
			argVector.add(cd.getSensedChemicals());
			System.out.println(cd.displaySensedChemicals());
			cell.provideService(argVector);
			cd.disposeSensedChemicals();
		}
		
		updateStatistics();
	}

	public Vector provideService(Vector args)
	{
		if(args == null)
			return null;
		String service = (String)args.elementAt(0);
		if(service.equals("Diffuse Chemicals"))
		{
			int id = ((Integer)args.elementAt(1)).intValue();
			Vector emissions = (Vector)args.elementAt(2);
			CellDetails cd = getCellDetails(id);
			Position pos = cd.getPosition();
			double x = pos.getX(), y = pos.getY(), z = pos.getZ();
			diffuseChemicals(x,y,z,emissions);
			
			Enumeration enum = emissions.elements();
			while(enum.hasMoreElements())
			{
				ChemicalConcentration cc = (ChemicalConcentration)enum.nextElement();
				System.out.println(cc.toString());
			}
		}
		else if(service.equals("Cell Division"))
		{
			int id = ((Integer)args.elementAt(1)).intValue();
			String state1 = (String)args.elementAt(2);
			String state2 = (String)args.elementAt(3);
			String divisionType = (String)args.elementAt(4);
			performCellDivision(id,state1,state2,divisionType);
		}
		else if(service.equals("Simulation Instruction"))
		{
			Vector instruction = (Vector)args.elementAt(1);
			runInstruction(instruction);
		}
		else if(service.equals("Get Statistics"))
		{
			Vector argVector = new Vector();
			argVector.add(statistics);
			return argVector;
		}
		return null;
	}

	void updateStatistics()
	{
		statistics.put(new String("Number of cells"), (new Integer(cellDetails.size())).toString());
	}

	public BranchGroup getDisplayGraph()
	{
		// Globally used colors
		Color3f black = new Color3f(0.0f, 0.0f, 0.0f);
		Color3f white = new Color3f(1.0f, 1.0f, 1.0f);

		BranchGroup objRoot = new BranchGroup();

		Enumeration enum = cellDetails.elements();
		int curr=0;
		while(enum.hasMoreElements())
		{
			CellDetails cd = (CellDetails)enum.nextElement();
			Cell cell = cd.getCell();

			Appearance appearance = new Appearance();
			// Set up the material properties
			Color3f objColor = cell.getColor().getColor3fObject();
			appearance.setMaterial(new Material(black, black, objColor,
					     white, 80.0f));

			TransformGroup objTrans = new TransformGroup();
			Transform3D t = new Transform3D();
			Position cellPos = cd.getPosition();
			t.set(cellPos.getVector3dObject());
			objTrans.setTransform(t);

			Node obj = new Box((float)cd.getRadius(),
								(float)cd.getRadius(),
								(float)cd.getRadius(),
								Box.GENERATE_NORMALS | Box.GENERATE_TEXTURE_COORDS, appearance);
			objTrans.addChild(obj);
			objRoot.addChild(objTrans);

			curr++;
		}

		return objRoot;
	}

	private CellDetails getCellDetails(int id)
	{
		Enumeration enum = cellDetails.elements();
		while(enum.hasMoreElements())
		{
			CellDetails cd = (CellDetails)enum.nextElement();

			if(id == cd.getCell().getID().intValue())
				return cd;
		}
		return null;
	}

	private void diffuseChemicals(double x1, double y1, double z1, Vector emissions)
	{
		double x2, y2, z2, dist;

		Enumeration enum = cellDetails.elements();
		while(enum.hasMoreElements())
		{
			CellDetails cd = (CellDetails)(enum.nextElement());
			Position destPos = cd.getPosition();
			x2 = destPos.getX();
			y2 = destPos.getY();
			z2 = destPos.getZ();
			dist = Math.sqrt( (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2) + (z1-z2)*(z1-z2) );

			Enumeration emissionEnum = emissions.elements();
			while(emissionEnum.hasMoreElements())
			{
				ChemicalConcentration emission = (ChemicalConcentration)emissionEnum.nextElement();
				double conc = emission.getConcentration()*Math.exp(-dist/2);
				String chem = emission.getChemical();
				ChemicalConcentration cc = new ChemicalConcentration(chem,conc);
				cd.addSensedChemical(cc);
			}
		}
	}

	private CellDetails getNewCell(String state)
	{
		CellDetails cd = null;
		Integer ID = new Integer(currCellID);

		try
		{
			Vector args = new Vector();
			args.add(new String(state));
			Cell cell = getSimulation().createCell(ID, this, cellProgram, args);
			cd = new CellDetails(cell);
			cellDetails.add(cd);
			currCellID++;
		}
		catch(Exception e)
		{
			System.err.println(e);
		}

		return cd;
	}

	private void performCellDivision(int id, String state1, String state2, String divisionType)
	{
		CellDetails cd1 = null, cd2;
		Enumeration enum = cellDetails.elements();
		while(enum.hasMoreElements())
		{
			CellDetails cd = (CellDetails)enum.nextElement();
			if(cd.getCell().getID().intValue() == id)
			{
				cd1 = cd;
				break;
			}
		}
		cd2 = getNewCell(state2);
		if(cd1 == null)
		{
			System.err.println("Unknown cell requested cell division");
			System.exit(0);
		}
		if(cd2 == null)
		{
			System.err.println("Unable to perform cell division");
			System.exit(0);
		}
		cd1.getCell().setState(state1);
		if(divisionType.equals("axis"))
			divideUsingAxisUnpolarised(cd1, cd2);
		else if(divisionType.equals("polarised"))
			divideUsingAxisPolarised(cd1, cd2);
		else if(divisionType.equals("normal-X"))
			divideUsingNormalX(cd1, cd2);
		else if(divisionType.equals("normal-Y"))
			divideUsingNormalY(cd1, cd2);
		cd2.setAsNew();
	}

	private void divideUsingNormalX (CellDetails cd1, CellDetails cd2)
	{
		double x = cd1.getPosition().getX();
		double y = cd1.getPosition().getY();
		double z = cd1.getPosition().getZ();
		double r = cd1.getRadius();
		double axis = normalAngle(cd1.getAxis());
		double normal = normalAngle(cd1.getNormalX());
		double axis1 = normalAngle(cd1.getAxis() - Math.PI/4.0);
		double axis2 = normalAngle(cd1.getAxis() + Math.PI/4.0);
		double normal1 = normalAngle(cd1.getNormalX() - Math.PI/4.0);
		double normal2 = normalAngle(cd1.getNormalX() + Math.PI/4.0);
		double dist = r;
		double temp;

		if(!approximatelySame(axis1 - Math.PI/2,normal1))
		{
			temp = axis1; axis1 = axis2; axis2 = temp;
			temp = normal1; normal1 = normal2; normal2 = temp;
		}

		cd2.setValues (x - dist/2 * Math.cos(normal), y - dist/2 * Math.sin(normal), z,
						Math.pow(r*r*r / 2,0.333333), axis2, normal2, 0.0);

		x = x + dist/2 * Math.cos(normal);
		y = y + dist/2 * Math.sin(normal);
		r = Math.pow(r*r*r / 2,0.333333);
		cd1.setValues(x,y,z,r,axis1,normal1, 0.0);
	}

	private void divideUsingNormalY (CellDetails cd1, CellDetails cd2)
	{
		double x = cd1.getPosition().getX();
		double y = cd1.getPosition().getY();
		double z = cd1.getPosition().getZ();
		double r = cd1.getRadius();
		double axis1 = normalAngle(cd1.getAxis() - Math.PI/4.0);
		double axis2 = normalAngle(cd1.getAxis() + Math.PI/4.0);
		double normal1 = normalAngle(cd1.getNormalX() - Math.PI/4.0);
		double normal2 = normalAngle(cd1.getNormalX() + Math.PI/4.0);
		double dist = r;
		double temp;

		if(!approximatelySame(axis1 - Math.PI/2,normal1))
		{
			temp = axis1; axis1 = axis2; axis2 = temp;
			temp = normal1; normal1 = normal2; normal2 = temp;
		}

		cd2.setValues (x, y, z - dist/2, Math.pow(r*r*r / 2,0.333333), axis2, normal2, 0.0);

		z += dist/2;
		r = Math.pow(r*r*r / 2,0.333333);
		cd1.setValues(x,y,z,r,axis1,normal1, 0.0);
	}

	private void divideUsingAxisPolarised (CellDetails cd1, CellDetails cd2)
	{
		double x = cd1.getPosition().getX();
		double y = cd1.getPosition().getY();
		double z = cd1.getPosition().getZ();
		double r = cd1.getRadius();
		double axis1 = normalAngle(cd1.getAxis());
		double normal1 = normalAngle(cd1.getNormalX());
		double normal2 = normalAngle(-cd1.getNormalX());
		double dist = r;

		cd2.setValues (x + dist/2 * Math.cos(axis1), y + dist/2 * Math.sin(axis1), z,
							Math.pow(r*r*r / 2,0.333333), axis1, normal2, 0.0);
		x = x - dist/2 * Math.cos(axis1);
		y = y - dist/2 * Math.sin(axis1);
		r = Math.pow(r*r*r / 2,0.333333);
		cd1.setValues(x,y,z,r,axis1,normal1, 0.0);
	}

	private void divideUsingAxisUnpolarised (CellDetails cd1, CellDetails cd2)
	{
		double x = cd1.getPosition().getX();
		double y = cd1.getPosition().getY();
		double z = cd1.getPosition().getZ();
		double r = cd1.getRadius();
		double axis1 = normalAngle(cd1.getAxis());
		double axis2 = normalAngle(cd1.getAxis() - Math.PI);
		double normal1 = normalAngle(cd1.getNormalX());
		double normal2 = normalAngle(-cd1.getNormalX() - Math.PI);
		double dist = r;

		axis2 = normalAngle(axis1 - Math.PI);
		normal2 = normalAngle(normal1 - Math.PI);
		cd2.setValues (x - dist/2 * Math.cos(axis1), y - dist/2 * Math.sin(axis1), z,
						Math.pow(r*r*r / 2,0.333333), axis2, normal2, 0.0);
		x = x + dist/2 * Math.cos(axis1);
		y = y + dist/2 * Math.sin(axis1);
		r = Math.pow(r*r*r / 2,0.333333);
		cd1.setValues(x,y,z,r,axis1,normal1, 0.0);
	}

	static double normalAngle(double angle)
	{
		if((angle > Math.PI)||(angle <= -Math.PI))
			angle -= 2*Math.PI*(int)( angle / (2*Math.PI) );
		if(angle == -Math.PI)
			angle = -angle;

		return angle;
	}

	static double positiveAngle(double angle)
	{
		if((angle>=2*Math.PI) || (angle<0))
			angle -= 2*Math.PI*(int)( angle / (2*Math.PI) );
		if(angle<0)
			angle -= 2*Math.PI*((int)( angle / (2*Math.PI) ) - 1);

		return angle;
	}

	static boolean approximatelySame(double angleA, double angleB)
	{
		if(positiveAngle(angleA - angleB) <= 0.01)
			return true;
		else if(positiveAngle(angleA - angleB) >= (Math.PI * 2 - 0.01))
			return true;
		else
			return false;
	}

	private void runInstruction(Vector instruction)
	{
		String instrType = (String)instruction.elementAt(0);
		if(instrType.equals("new"))
		{
			CellDetails cd = getNewCell((String)instruction.elementAt(1));
			cd.setValues(((Double)instruction.elementAt(2)).doubleValue(),
							((Double)instruction.elementAt(3)).doubleValue(),
							((Double)instruction.elementAt(4)).doubleValue(),
							((Double)instruction.elementAt(5)).doubleValue(),
							((Double)instruction.elementAt(6)).doubleValue(),
							((Double)instruction.elementAt(7)).doubleValue(),
							((Double)instruction.elementAt(8)).doubleValue());
		}
		else if(instrType.equals("kill"))
		{
			double x1, y1, z1, x2, y2, z2, radius, dist;
			x1 = ((Double)instruction.elementAt(1)).doubleValue();
			y1 = ((Double)instruction.elementAt(2)).doubleValue();
			z1 = ((Double)instruction.elementAt(3)).doubleValue();
			radius = ((Double)instruction.elementAt(3)).doubleValue();
			Enumeration enum = cellDetails.elements();
			while(enum.hasMoreElements())
			{
				CellDetails cd = (CellDetails)(enum.nextElement());
				Position destPos = cd.getPosition();
				x2 = destPos.getX();
				y2 = destPos.getY();
				z2 = destPos.getZ();
				dist = Math.sqrt( (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2) + (z1-z2)*(z1-z2) );
				if(dist <= radius)
				{
					cellDetails.remove(cd);
				}
			}
		}
	}

	class CellDetails
	{
		Cell cell;
		Position position;
		double radius;
		double axis;
		double normalX;
		double normalY;
		Vector sensedChemicals;
		boolean isNewCell;

		CellDetails(Cell p_cell) { cell = p_cell; sensedChemicals = new Vector(); isNewCell = false; }
		public Cell getCell() { return cell; }
		public Position getPosition() { return new Position(position); }
		public void setPosition(Position p_pos) { position = p_pos; }
		public double getRadius() { return radius; }
		public double getAxis() { return axis; }
		public double getNormalX() { return normalX; }
		public double getNormalY() { return normalY; }
		public void setValues(double p_x, double p_y, double p_z, double p_radius,
								double p_axis, double p_normalX, double p_normalY)
		{
			position = new Position(p_x, p_y, p_z);
			radius = p_radius;
			axis = p_axis;
			normalX = p_normalX;
			normalY = p_normalY;
		}
		public void addSensedChemical(ChemicalConcentration emission)
		{
			boolean chemWasFound = false;
			Enumeration enum = sensedChemicals.elements();
			while(enum.hasMoreElements())
			{
				ChemicalConcentration cc = (ChemicalConcentration) enum.nextElement();
				if(emission.getChemical().equals(cc.getChemical()))
				{
					cc.setConcentration(cc.getConcentration()+emission.getConcentration());
					chemWasFound = true;
					break;
				}
			}
			if(!chemWasFound)
			{
				sensedChemicals.add(new ChemicalConcentration(emission));
			}
		}
		public Vector getSensedChemicals() { return sensedChemicals; }
		public void disposeSensedChemicals() { sensedChemicals = new Vector(); }
		public String displaySensedChemicals()
		{
			Enumeration enum = sensedChemicals.elements();
			StringBuffer sb = new StringBuffer();
			sb.append(""+sensedChemicals.size());
			while(enum.hasMoreElements())
			{
				ChemicalConcentration cc = (ChemicalConcentration)enum.nextElement();
				sb.append(cc.toString() + ",");
			}
			return sb.toString();
		}
		public boolean isNew() { return isNewCell; }
		public void setAsNew() { isNewCell = true; }
		public void setAsOld() { isNewCell = false; }
	}

	static class CoordinateGeometryFactory extends SystemFactory
	{
		public SystemModel create(Simulation p_sim, String modelOptions)
		{
			CoordinateGeometry sb = new CoordinateGeometry(p_sim);

			Vector vec = new Vector();
			vec.add(modelOptions);
			sb.init(vec);

			return sb;
		}
	}

	static
	{
		Simulation.sysModels.put(new String("sysmodels.coordinategeometry.CoordinateGeometry"),new CoordinateGeometryFactory());
	}
}
