package sysmodels.turingmachine;

import java.util.*;
import javax.media.j3d.*;
import javax.vecmath.*;
import com.sun.j3d.utils.geometry.Sphere;

import core.*;
import util.*;

class TuringMachine extends SystemModel
{
	CellProgram cellProgram;
	Vector cellDetails;
	int currCellID;

	public TuringMachine(Simulation p_sim)
	{
		super(p_sim);
		currCellID = 0;
	}

	public void init(Vector args)
	{
		String modelOptions = (String) args.elementAt(0);

		cellProgram = getSimulation().getCellProgram();

		cellDetails = new Vector();

		CellDetails cd1 = getNewCell(new String("emitA_s1"));
		cd1.setValues(-1.5, 0.0, 0.0, 0.25);
		CellDetails cd2 = getNewCell(new String("emitA"));
		cd2.setValues(-1.0, 0.0, 0.0, 0.25);
		CellDetails cd3 = getNewCell(new String("emitB"));
		cd3.setValues(-0.5, 0.0, 0.0, 0.25);
		CellDetails cd4 = getNewCell(new String("emitB"));
		cd4.setValues(0.0, 0.0, 0.0, 0.25);
		CellDetails cd5 = getNewCell(new String("emitC"));
		cd5.setValues(0.5, 0.0, 0.0, 0.25);
		CellDetails cd6 = getNewCell(new String("emitC"));
		cd6.setValues(1.0, 0.0, 0.0, 0.25);
		CellDetails cd7 = getNewCell(new String("emitBlank"));
		cd7.setValues(1.5, 0.0, 0.0, 0.25);

		Vector emissions = new Vector();
		emissions.add(new ChemicalConcentration("a",1.0));
		diffuseChemicals(cd1.getCell().getID().intValue(), emissions);
		emissions = new Vector();
		emissions.add(new ChemicalConcentration("a",1.0));
		diffuseChemicals(cd2.getCell().getID().intValue(), emissions);
		emissions = new Vector();
		emissions.add(new ChemicalConcentration("b",1.0));
		diffuseChemicals(cd3.getCell().getID().intValue(), emissions);
		emissions = new Vector();
		emissions.add(new ChemicalConcentration("b",1.0));
		diffuseChemicals(cd4.getCell().getID().intValue(), emissions);
		emissions = new Vector();
		emissions.add(new ChemicalConcentration("c",1.0));
		diffuseChemicals(cd5.getCell().getID().intValue(), emissions);
		emissions = new Vector();
		emissions.add(new ChemicalConcentration("c",1.0));
		diffuseChemicals(cd6.getCell().getID().intValue(), emissions);

		Enumeration enum = cellDetails.elements();
		while(enum.hasMoreElements())
		{
			Vector argVector = new Vector();
			CellDetails cd = (CellDetails)enum.nextElement();
			Cell cell = cd.getCell();
			argVector.add(new String("Sense Chemicals"));
			argVector.add(cd.getSensedChemicals());
			cell.provideService(argVector);
			cd.disposeSensedChemicals();
		}
	}

	public void runCellProgramForOneStep()
	{
		Enumeration enum = cellDetails.elements();
		while(enum.hasMoreElements())
		{
			CellDetails cd = (CellDetails)enum.nextElement();
			if(cd.isNew())
			{
				cd.setAsOld();
			}
			else
			{
				Cell cell = cd.getCell();
				cell.runCellProgramForOneStep();
			}
		}

		/* Now distribute the chemicals diffused during the run to individual cells */
		enum = cellDetails.elements();
		while(enum.hasMoreElements())
		{
			Vector argVector = new Vector();
			CellDetails cd = (CellDetails)enum.nextElement();
			Cell cell = cd.getCell();
			argVector.add(new String("Sense Chemicals"));
			argVector.add(cd.getSensedChemicals());
			cell.provideService(argVector);
			cd.disposeSensedChemicals();
		}
	}

	public Vector provideService(Vector args)
	{
		if(args == null)
			return null;
		String service = (String)args.elementAt(0);
		if(service.equals("Diffuse Chemicals"))
		{
			int id = ((Integer)args.elementAt(1)).intValue();
			Vector emissions = (Vector)args.elementAt(2);
			diffuseChemicals(id,emissions);
		}
		else if(service.equals("Cell Division"))
		{
			int id = ((Integer)args.elementAt(1)).intValue();
			String state1 = (String)args.elementAt(2);
			String state2 = (String)args.elementAt(3);
			String divisionType = (String)args.elementAt(4);
			performCellDivision(id,state1,state2,divisionType);
		}
		else if(service.equals("Simulation Instruction"))
		{
			/*Vector strings = (Vector)args.elementAt(1);
			SimInstruction instruction = new SimInstruction(strings);
			executeSimulationInstruction(instruction);*/
		}
		return null;
	}

	public BranchGroup getDisplayGraph()
	{
		// Globally used colors
		Color3f black = new Color3f(0.0f, 0.0f, 0.0f);
		Color3f white = new Color3f(1.0f, 1.0f, 1.0f);

		BranchGroup objRoot = new BranchGroup();

		Enumeration enum = cellDetails.elements();
		int curr=0;
		while(enum.hasMoreElements())
		{
			CellDetails cd = (CellDetails)enum.nextElement();
			Cell cell = cd.getCell();

			Appearance appearance = new Appearance();
			// Set up the material properties
			Color3f objColor = cell.getColor().getColor3fObject();
			appearance.setMaterial(new Material(black, black, objColor,
					     white, 80.0f));

			TransformGroup objTrans = new TransformGroup();
			Transform3D t = new Transform3D();
			Position cellPos = cd.getPosition();
			t.set(cellPos.getVector3dObject());
			objTrans.setTransform(t);

			Node obj = new Sphere((float)cd.getRadius(), Sphere.GENERATE_NORMALS | Sphere.GENERATE_TEXTURE_COORDS, 20, appearance);
			objTrans.addChild(obj);
			objRoot.addChild(objTrans);

			curr++;
		}

		return objRoot;
	}

	private void diffuseChemicals(int id, Vector emissions)
	{
		Position srcPos, destPos;
		double x1, y1, z1, x2, y2, z2, r, dist, conc;
		String chem;

		srcPos = null; x1 = 0.0; y1 = 0.0; z1 = 0.0; r = 1.0;

		Enumeration enum = cellDetails.elements();
		while(enum.hasMoreElements())
		{
			CellDetails cd = (CellDetails)enum.nextElement();
			
			if(id == cd.getCell().getID().intValue())
			{
				srcPos = cd.getPosition();
				x1 = srcPos.getX();
				y1 = srcPos.getY();
				z1 = srcPos.getZ();
				r = cd.getRadius();
				break;
			}
		}

		if(srcPos == null)
		{
			System.err.println("Cell with unknown ID");
			System.exit(0);
		}

		enum = cellDetails.elements();
		while(enum.hasMoreElements())
		{
			CellDetails cd = (CellDetails)(enum.nextElement());
			destPos = cd.getPosition();
			x2 = destPos.getX();
			y2 = destPos.getY();
			z2 = destPos.getZ();
			dist = Math.sqrt( (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2) + (z1-z2)*(z1-z2) );

			Enumeration emissionEnum = emissions.elements();
			while(emissionEnum.hasMoreElements())
			{
				ChemicalConcentration emission = (ChemicalConcentration)emissionEnum.nextElement();
				conc = emission.getConcentration() - dist/(4*r);
				if(conc < 0)
					conc = 0.0;
				chem = emission.getChemical();
				ChemicalConcentration cc = new ChemicalConcentration(chem,conc);
				cd.addSensedChemical(cc);
			}
		}
	}

	private CellDetails getNewCell(String state)
	{
		CellDetails cd = null;
		Integer ID = new Integer(currCellID);

		try
		{
			Vector args = new Vector();
			args.add(new String(state));
			Cell cell = getSimulation().createCell(ID, this, cellProgram, args);
			cd = new CellDetails(cell);
			cellDetails.add(cd);
			currCellID++;
		}
		catch(Exception e)
		{
			System.err.println(e);
		}

		return cd;
	}

	private void performCellDivision(int id, String state1, String state2, String divisionType)
	{
		CellDetails cd1 = null, cd2;
		Enumeration enum = cellDetails.elements();
		while(enum.hasMoreElements())
		{
			CellDetails cd = (CellDetails)enum.nextElement();
			if(cd.getCell().getID().intValue() == id)
			{
				cd1 = cd;
				break;
			}
		}
		cd2 = getNewCell(state2);
		if(cd1 == null)
		{
			System.err.println("Unknown cell requested cell division");
			System.exit(0);
		}
		if(cd2 == null)
		{
			System.err.println("Unable to perform cell division");
			System.exit(0);
		}

		cd1.getCell().setState(state1);
		divideCell(cd1, cd2);
		cd2.setAsNew();
	}

	private void divideCell (CellDetails cd1, CellDetails cd2)
	{
		double x = cd1.getPosition().getX();
		double y = cd1.getPosition().getY();
		double z = cd1.getPosition().getZ();
		double r = cd1.getRadius();
		double dist = 2*r;

		cd2.setValues (x + dist, y, z, r);
	}

	class CellDetails
	{
		Cell cell;
		Position position;
		double radius;
		Vector sensedChemicals;
		boolean isNewCell;

		CellDetails(Cell p_cell) { cell = p_cell; sensedChemicals = new Vector(); isNewCell = false; }
		public Cell getCell() { return cell; }
		public Position getPosition() { return new Position(position); }
		public void setPosition(Position p_pos) { position = p_pos; }
		public double getRadius() { return radius; }
		public void setRadius(double pRadius) { radius = pRadius; }
		public void setValues(double p_x, double p_y, double p_z, double p_radius)
		{
			position = new Position(p_x, p_y, p_z);
			radius = p_radius;
		}
		public void addSensedChemical(ChemicalConcentration emission)
		{
			boolean chemWasFound = false;
			Enumeration enum = sensedChemicals.elements();
			while(enum.hasMoreElements())
			{
				ChemicalConcentration cc = (ChemicalConcentration) enum.nextElement();
				if(emission.getChemical().equals(cc.getChemical()))
				{
					cc.setConcentration(cc.getConcentration()+emission.getConcentration());
					chemWasFound = true;
					break;
				}
			}
			if(!chemWasFound)
			{
				sensedChemicals.add(new ChemicalConcentration(emission));
			}
		}
		public Vector getSensedChemicals() { return sensedChemicals; }
		public void disposeSensedChemicals() { sensedChemicals = new Vector(); }
		public boolean isNew() { return isNewCell; }
		public void setAsNew() { isNewCell = true; }
		public void setAsOld() { isNewCell = false; }
	}

	class SimInstruction
	{
		public final int NOOP = 0;
		public final int NEW = 1;
		public final int DIFFUSE = 2;
		public int type;
		public String name;
		public Position pos;
		public float quantity;
		
		SimInstruction(Vector strings)
		{
			/*String str = strings.elementAt(0);
			if(str.equals("new")) { type = NEW; }
			else if(str.equals("diffuse"))) { type = DIFFUSE; }
			else { type = NOOP; }
			name = strings.elementAt(1);*/
		}
	}

	static class TuringMachineFactory extends SystemFactory
	{
		public SystemModel create(Simulation p_sim, String modelOptions)
		{
			TuringMachine sb = new TuringMachine(p_sim);

			Vector vec = new Vector();
			vec.add(modelOptions);
			sb.init(vec);

			return sb;
		}
	}

	static
	{
		Simulation.sysModels.put(new String("sysmodels.turingmachine.TuringMachine"),new TuringMachineFactory());
	}
}
